@@ -2,9 +2,11 @@
 """A generic, resizable thread pool"""
 
 from Queue import Queue
+from itertools import count
 from threading import Lock, Thread, current_thread
 
 from application import log
+from application.python import limit
 from application.python.decorator import decorator, preserve_signature
 
 
@@ -24,26 +26,43 @@ class ThreadPool(object):
     StopWorker = object()
 
     def __init__(self, name=None, min_threads=1, max_threads=10):
+        assert 0 <= min_threads <= max_threads > 0, "invalid bounds"
         self.name = name
         self._lock = Lock()
         self._queue = Queue()
+        self._thread_id = None
         self._threads = []
-        self._workers = 0
         self._started = False
-        self._set_size(min_threads, max_threads)
+        self.__dict__['min_threads'] = min_threads
+        self.__dict__['max_threads'] = max_threads
+        self.__dict__['workers'] = 0
+        self.__dict__['jobs'] = 0
 
+    @property
     def min_threads(self):
         return self.__dict__['min_threads']
 
+    @property
     def max_threads(self):
         return self.__dict__['max_threads']
 
+    @property
+    def workers(self):
+        return self.__dict__['workers']
+
+    @property
+    def jobs(self):
+        return self.__dict__['jobs']
+
     def start(self):
         with self._lock:
             if self._started:
                 return
             self._started = True
-            self._set_size(self.min_threads, self.max_threads)
+            self._thread_id = count(1)
+            needed_workers = limit(self.jobs, min=self.min_threads, max=self.max_threads)
+            while self.workers < needed_workers:
+                self._start_worker()
 
     def stop(self):
         with self._lock:
@@ -51,47 +70,41 @@ class ThreadPool(object):
                 return
             self._started = False
             threads = self._threads[:]
-            while self._workers:
+            while self.workers:
                 self._stop_worker()
             for thread in threads:
                 thread.join()
+            self._thread_id = None
 
     def resize(self, min_threads=1, max_threads=10):
+        assert 0 <= min_threads <= max_threads > 0, "invalid bounds"
         with self._lock:
-            self._set_size(min_threads, max_threads)
+            self.__dict__['min_threads'] = min_threads
+            self.__dict__['max_threads'] = max_threads
+            if self._started:
+                needed_workers = limit(self.jobs, min=min_threads, max=max_threads)
+                while self.workers > max_threads:  # compare against needed_workers to compact or against max_threads to not
+                    self._stop_worker()
+                while self.workers < needed_workers:
+                    self._start_worker()
 
     def compact(self):
         with self._lock:
-            while self._workers > self.min_threads:
+            needed_workers = limit(self.jobs, min=self.min_threads, max=self.max_threads)
+            while self.workers > needed_workers:
                 self._stop_worker()
 
     def run(self, func, *args, **kw):
         with self._lock:
             self._queue.put(CallFunctionEvent(func, args, kw))
-            if self._started:
-                self._maybe_start_workers()
-
-    def _set_size(self, min, max):
-        # Must be called with the lock held
-        assert 0 <= min <= max, "invalid bounds"
-
-        self.__dict__['min_threads'] = min
-        self.__dict__['max_threads'] = max
-
-        if not self._started:
-            return
-
-        while self._workers > max:
-            self._stop_worker()
-        while self._workers < min:
-            self._start_worker()
-
-        self._maybe_start_workers()
+            self.__dict__['jobs'] += 1
+            if self._started and self.workers < limit(self.jobs, max=self.max_threads):
+                self._start_worker()
 
     def _start_worker(self):
         # Must be called with the lock held
-        self._workers += 1
-        name = "%sThread-%s-%s" % (self.__class__.__name__, self.name or id(self), self._workers)
+        self.__dict__['workers'] += 1
+        name = "%sThread-%s-%s" % (self.__class__.__name__, self.name or id(self), next(self._thread_id))
         thread = Thread(target=self._worker, name=name)
         self._threads.append(thread)
         thread.daemon = True
@@ -100,26 +113,22 @@ class ThreadPool(object):
     def _stop_worker(self):
         # Must be called with the lock held
         self._queue.put(self.StopWorker)
-        self._workers -= 1
-
-    def _maybe_start_workers(self):
-        # Must be called with the lock held
-        needed = self._queue.qsize() + self._workers
-        while self._workers < min(self.max_threads, needed):
-            self._start_worker()
+        self.__dict__['workers'] -= 1
 
     def _worker(self):
         thread = current_thread()
         while True:
-            item = self._queue.get()
-            if item is self.StopWorker:
+            task = self._queue.get()
+            if task is self.StopWorker:
                 break
             try:
-                item.function(*item.args, **item.kw)
+                task.function(*task.args, **task.kw)
             except:
-                log.error('Exception occurred while calling %r in the %r thread' % (item.function, thread.name))
-                log.err()
-            del item
+                log.exception('Exception occurred while calling %r in the %r thread' % (task.function, thread.name))
+            finally:
+                with self._lock:
+                    self.__dict__['jobs'] -= 1
+                del task
         self._threads.remove(thread)
 
 
